/*
 * Copyright 2013 Vincent Oostindië
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License
 */

package nl.ulso.sprox.impl;

import nl.ulso.sprox.Namespace;
import nl.ulso.sprox.Namespaces;

import javax.xml.namespace.QName;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import static java.util.Collections.emptyMap;
import static javax.xml.XMLConstants.NULL_NS_URI;

/**
 * Represents a controller class.
 * <p/>
 * On construction all namespaces (if any) are pulled from from the class and verified. All {@link ControllerMethod}s
 * for the same class share the same instance of this class, providing access to the namespaces.
 */
final class ControllerClass<T> {

    private final Class<T> clazz;
    private  Map<String, String> namespaces;
    private String defaultNamespace;

    static <T> ControllerClass<T> createControllerClass(Class<T> clazz) {
        final ControllerClass<T> controllerClass = new ControllerClass<>(clazz);
        controllerClass.initialize();
        return controllerClass;
    }

    private ControllerClass(Class<T> clazz) {
        this.clazz = clazz;
    }

    // A separate initialization step is necessary because otherwise "this" would escape the constructor.
    private void initialize() {
        this.namespaces = extractNamespaces();
        this.defaultNamespace = determineDefaultNamespace();
    }

    private Map<String, String> extractNamespaces() {
        final Namespaces namespacesAnnotation = clazz.getAnnotation(Namespaces.class);
        final Namespace namespaceAnnotation = clazz.getAnnotation(Namespace.class);
        if (namespacesAnnotation != null && namespaceAnnotation != null) {
            throw new IllegalStateException("Controller class '" + clazz
                    + "' must have either no annotations, a @Namespace or a @Namespaces annotation. This one has both.");
        }
        if (namespacesAnnotation != null) {
            return extractMultipleNamespaces(namespacesAnnotation);
        } else if (namespaceAnnotation != null) {
            return extractSingleNamespace(namespaceAnnotation);
        }
        return emptyMap();
    }

    private Map<String, String> extractMultipleNamespaces(Namespaces namespacesAnnotation) {
        Map<String, String> result = new HashMap<>(namespacesAnnotation.value().length);
        for (Namespace namespace : namespacesAnnotation.value()) {
            final String shorthand = namespace.shorthand();
            final String name = namespace.value();
            if (result.containsKey(shorthand)) {
                throw new IllegalStateException("Duplicate shorthands are not allowed: " + shorthand);
            }
            result.put(shorthand, name);
        }
        return result;
    }

    private Map<String, String> extractSingleNamespace(Namespace namespaceAnnotation) {
        Map<String, String> result = new HashMap<>(1);
        result.put(namespaceAnnotation.shorthand(), namespaceAnnotation.value());
        return result;
    }

    private String determineDefaultNamespace() {
        if (namespaces.isEmpty()) {
            return null;
        }
        if (clazz.isAnnotationPresent(Namespace.class)) {
            return clazz.getAnnotation(Namespace.class).value();
        }
        return clazz.getAnnotation(Namespaces.class).value()[0].value();
    }

    /**
     * Invokes the specified method on the controller of this class in the execution context, passing it the method
     * parameters.
     *
     * @param method Method to invoke.
     * @param context Context to pull the controller for this class from.
     * @param methodParameters Parameters to pass to the controller.
     * @return The result of invoking the method.
     * @throws IllegalStateException If the method could not be invoked.
     */
    Object invokeMethod(Method method, ExecutionContext context, Object[] methodParameters) {
        try {
            return method.invoke(context.getController(clazz), methodParameters);
        } catch (IllegalAccessException | InvocationTargetException e) {
            throw new IllegalStateException("Could not invoke method '" + method
                    + "' on controller of class '" + clazz + "'");
        }
    }

    /**
     * Constructs a QName from a reference (an annotation value) using the default namespace as set on the class (if
     * any) if none was defined in the reference itself.
     *
     * @param reference Value to construct the QName for.
     * @return A new QName
     */
    QName createQName(String reference) {
        return createQName(reference, defaultNamespace);
    }

    /**
     * Constructs a QName from a reference (an annotation value), falling back on the specific default namespace if none
     * could be derived from the reference.
     * <p/>
     * A reference can be of the form {@code "&lt;elementName&gt;"} or {@code "&lt;shorthand&gt;:&lt;elementName&gt;"}.
     * In the latter case, the shorthand refers to a namespace specified on the class.
     *
     * @param reference Value to construct the QName for.
     * @return A new QName
     */
    QName createQName(String reference, String defaultNamespace) {
        final String localPart;
        final String namespace;
        final int i = reference.indexOf(':');
        if (i == -1) {
            localPart = reference;
            namespace = defaultNamespace;
        } else {
            localPart = reference.substring(i + 1);
            namespace = namespaces.get(reference.substring(0, i));
        }
        return new QName(namespace == null ? NULL_NS_URI : namespace, localPart);
    }
}
